import os
import copy
from collections import OrderedDict

import numpy as np

from ...Workflow import Workflow 
from ...IO.BL15 import ReadTimeSeries 
from .. import IntegrateBatch

inputs = copy.deepcopy(ReadTimeSeries.inputs)
inputs.update(
    integrator = None,
    q_min = 0.,
    q_max = float('inf'),
    output_dir = None
    )

outputs = copy.deepcopy(IntegrateBatch.outputs)

class IntegrateTimeSeries(Workflow):
    
    def __init__(self):
        super(IntegrateTimeSeries,self).__init__(inputs,outputs)
        self.add_operations(
            read = ReadTimeSeries.ReadTimeSeries(),
            integrate = IntegrateBatch.IntegrateBatch()
            )

    def run(self):
        self.operations['read'].operations['read_batch'].operations['read'].disable_ops('read_q_I','read_system')
        read_inputs = OrderedDict([(k,self.inputs[k]) for k in ReadTimeSeries.inputs.keys()])
        read_outputs = self.operations['read'].run_with(**read_inputs)

        self.outputs = self.operations['integrate'].run_with(
            image_data = read_outputs['image_data'],
            integrator = self.inputs['integrator'],
            q_min = self.inputs['q_min'],
            q_max = self.inputs['q_max'],
            output_dir = self.inputs['output_dir'],
            output_filenames = read_outputs['filenames']
            )

        return self.outputs




import copy
from collections import OrderedDict

from paws.workflows.Workflow import Workflow 

from paws.workflows.IO.BL15 import ReadTimeSeries
from paws.operations.VISUALIZATION.XRSDPlot import SortBatch

inputs = copy.deepcopy(ReadTimeSeries.inputs)
inputs.update(
    source_wavelength = None,
    show_plots = False
    )

outputs = OrderedDict(
    figures = [] 
    ) 

class PlotTimeSeries(Workflow):

    def __init__(self):
        self.add_operations(
            read = ReadTimeSeries.ReadTimeSeries(),
            plot = XRSDPlot()
            )

    def run(self):
        self.outputs = copy.deepcopy(outputs)
        self.operations['read'].operations['read_batch'].operations['read'].disable_ops('read_image')
        read_inputs = OrderedDict([(k,self.inputs[k]) for k in ReadTimeSeries.inputs.keys()])
        read_outputs = self.operations['read'].run_with(**read_inputs)

        for q_I, dI, sys in zip(read_outputs['q_I'],read_outputs['dI'],read_outputs['system']):
            plot_outputs = self.operations['plot'].run_with(
                q_I = q_I,
                source_wavelength = self.inputs['source_wavelength'],
                dI = dI,
                system = sys,
                show_plot = self.inputs['show_plots']
            )
            self.outputs['figures'].append(plot_outputs['figure'])
        return self.outputs




import os
from collections import OrderedDict

from xrsdkit import system as xrsdsys
from xrsdkit.tools import ymltools as xrsdyml

from paws.workflows.Workflow import Workflow 
from paws.operations.IO.NumpyLoad import NumpyLoad 
from paws.operations.IO.YAML.SaveXRSDSystem import SaveXRSDSystem
from paws.operations.PROCESSING.FITTING.XRSDFit import XRSDFit as XRSDFit_op

inputs = OrderedDict(
    q_I_file = None,
    system_file = None,
    q_I = None,
    system = None,
    error_weighted = True,
    logI_weighted = True,
    q_range = [0.,float('inf')], 
    experiment_id = None,
    sample_id = None,
    data_file_path = None,
    output_file = None
    )

outputs = OrderedDict(
    system_opt = None,
    system_opt_dict = None,
    q_I_opt = None,
    fit_report = None
    )

class XRSDFit(Workflow):

    def __init__(self):
        super(XRSDFit,self).__init__(inputs,outputs)
        self.add_operations(
            read_q_I = NumpyLoad(),
            fit = XRSDFit_op(),
            save_system = SaveXRSDSystem()
        )

    def run(self):
        q_I = self.inputs['q_I']
        if self.inputs['q_I_file']:
            read_outputs = self.operations['read_q_I'].run_with(file_path=self.inputs['q_I_file'])
            q_I = read_outputs['data']
        sys = self.inputs['system']
        if self.inputs['system_file']:
            sys = xrsdyml.load_sys_from_yaml(self.inputs['system_file'])
        if not sys: sys = xrsdsys.System()
        sys.sample_metadata['experiment_id'] = self.inputs['experiment_id']
        sys.sample_metadata['sample_id'] = self.inputs['sample_id']
        fit_outputs = self.operations['fit'].run_with(
            q_I = q_I,
            system = sys,
            error_weighted = self.inputs['error_weighted'],
            logI_weighted = self.inputs['logI_weighted'],
            q_range = self.inputs['q_range']
            )
        self.outputs['system_opt'] = fit_outputs['system']
        self.outputs['system_opt_dict'] = fit_outputs['system_dict']
        self.outputs['q_I_opt'] = fit_outputs['q_I_opt']
        self.outputs['fit_report'] = fit_outputs['system'].fit_report
        fit_sys = fit_outputs['system']
        fit_sys.sample_metadata.update(dict(
            experiment_id = self.inputs['experiment_id'],
            sample_id = self.inputs['sample_id'],
            data_file = self.inputs['data_file_path']
            ))
        if self.inputs['output_file']:
            self.operations['save_system'].run_with(
                system = fit_sys, 
                file_path = self.inputs['output_file']
                )
        return self.outputs       
 



from __future__ import print_function
from collections import OrderedDict
import os
import copy
import time

from .PawsPlugin import PawsPlugin

content = OrderedDict(
    timer=None,
    cryocon=None,
    ppumps={},
    volume_limits={},
    bad_flow_tolerance=1E9)

class FlowReactor(PawsPlugin):

    def __init__(self):
        super(FlowReactor,self).__init__(content)
        self.thread_blocking = True
        self.content_doc['timer'] = 'A Timer plugin for triggering plugin activities'
        self.content_doc['cryocon'] = 'CryoConController plugin for temperature control' 
        self.content_doc['ppumps'] = 'Dict of named MitosPPumpControllers' 
        self.content_doc['volume_limits'] = 'Dict of pump volume limits (in microlitres)' 
        #self.content_doc['solvent_pump_name'] = 'Name (string) of the pump controlling solvent flow' 
        self.content_doc['bad_flow_tolerance'] = 'Number of timer ticks '\
            'to tolerate bad flow readings before stopping the reactor' 
        self.thread_blocking = True
        self.verbose = False
        self.anomaly_count = 0
        self.solvent_name = None

    def description(self):
        desc = 'FlowReactor Plugin: '\
            'Coordinates a CryoConController '\
            'and a set of MitosPPumpControllers '\
            'to set recipes for a flow reactor.'
        return desc

    def start(self):
        super(FlowReactor,self).start()

    def run(self):
        tmr = self.content['timer'] 
        ppcs = self.content['ppumps']
        cryo = self.content['cryocon']
        bftol = self.content['bad_flow_tolerance']
        vlimits = self.content['volume_limits']
        vdelivered = dict.fromkeys(vlimits.keys())
        for k in vdelivered.keys(): vdelivered[k] = 0.
        with tmr.dt_lock:
            t_now = tmr.dt_utc()
        with self.proxy.history_lock:
            self.proxy.add_to_history(t_now,'START')
        keep_going = True
        stop_flag = False
        timer_dt_minutes = float(tmr.content['dt'])/60.
        #for nm,ppc in ppcs.items():
        #    ppc.set_flowrate(0.)
        if cryo:
            for chan in cryo.content['channels'].keys():
                cryo.set_units(chan,'C')
                cryo.loop_setup(chan,'RAMPP')
            cryo.start_control()
        while keep_going: 
            with tmr.dt_lock:
                tmr.dt_lock.wait()
            # log the flow rates and temperature,
            # check for anomalies in flow rates,
            # stop the reactor if anomalies found
            ok_flag,statstr,statdict = self.check_status()
            if ok_flag:
                self.anomaly_count = 0
            else:
                self.anomaly_count += 1
                if self.anomaly_count >= bftol:
                    if self.verbose: self.message_callback('Anomalous flow detected: stopping FlowReactor')
                    stop_flag = True
                
            if vlimits:
                for pump_nm, lmt in vlimits.items():
                    # get the true ul/m flowrate from statdict
                    fr = statdict['{}_flowrate'.format(pump_nm)]
                    # add flowrate * dt in minutes vdelivered[pump_nm]
                    vdelivered[pump_nm] += fr*timer_dt_minutes
                    # check if vdelivered[pump_nm] >= lmt: if so, throw the stop.
                    #if self.verbose: self.message_callback(
                    #    '{} volume delivered: {} / {}'.format(
                    #    pump_nm,vdelivered[pump_nm],lmt))
                    if vdelivered[pump_nm] >= lmt:
                        stop_flag = True
                        if self.verbose: self.message_callback('Volume limit reached: stopping FlowReactor')

            if stop_flag:
                with self.proxy.running_lock:
                    self.proxy.stop()
             
            with tmr.running_lock:
                if not tmr.running:
                    with self.proxy.running_lock:
                        self.proxy.stop()
            with self.proxy.running_lock:
                keep_going = bool(self.proxy.running)
        with tmr.dt_lock:
            t_now = tmr.dt_utc()
        if self.verbose: self.message_callback('Stopping FlowReactor at {}'.format(t_now))
        with self.proxy.history_lock:
            self.proxy.add_to_history(t_now,'STOP')
            self.proxy.dump_history()
        # set pumps to idle, stop cryocon loop
        for nm,ppc in ppcs.items():
            if self.verbose: self.message_callback('Setting pump {} to idle'.format(nm))
            ppc.set_idle()
        if cryo:
            if self.verbose: self.message_callback('Stopping CryoCon control loop')
            cryo.stop_control() 

    def set_recipe(self,recipe):
        cryo = self.content['cryocon']
        tmr = self.content['timer']
        self.solvent_name = recipe['solvent']
        for chan,loop_idx in cryo.content['channels'].items():
            if loop_idx is not None:
                if 'T_ramp' in recipe:
                    cryo.set_ramp_rate(chan,recipe['T_ramp'])
                if 'T_set' in recipe:
                    with cryo.state_lock:
                        T_current = cryo.state['T_read_{}'.format(chan)]
                    # NOTE: this is a little dance to circumvent a problem with the CryoCon,
                    # where it sometimes ignores the ramp rate 
                    cryo.set_temperature(chan,T_current)
                    T_inter = T_current + (recipe['T_set']-T_current)*0.1
                    cryo.set_temperature(chan,T_inter)
                    time.sleep(5)
                    cryo.set_temperature(chan,recipe['T_set'])
        ppcs = self.content['ppumps']
        if 'flowrate' in recipe:
            tot_frt = recipe['flowrate']
            solv_frt = tot_frt 
        for itm_nm, val in recipe.items():
            if '_fraction' in itm_nm:
                rg_nm = itm_nm[:itm_nm.find('_fraction')]
                rg_frt = tot_frt * val
                solv_frt -= rg_frt
                if self.verbose: self.message_callback('setting {} to {}/min'.format(rg_nm,rg_frt))
                ppcs[rg_nm].set_flowrate(rg_frt)
        if self.verbose: self.message_callback('setting solvent ({}) to {}/min'.format(self.solvent_name,solv_frt))
        ppcs[self.solvent_name].set_flowrate(solv_frt)
        if self.verbose: self.message_callback(self.prettyprint_recipe(recipe))
        with tmr.dt_lock:
            t_now = float(tmr.dt_utc())
        with self.history_lock:
            self.add_to_history(t_now,self.print_recipe(recipe))

    @staticmethod
    def prettyprint_recipe(rcp):
        rcp_str = ''
        if 'T_set' in rcp:
            rcp_str += os.linesep+'T_set: {}'.format(rcp['T_set'])
        if 'T_ramp' in rcp: 
            rcp_str += os.linesep+'T_ramp: {}/min'.format(rcp['T_ramp'])
        if 'flowrate' in rcp:
            tot_frt = rcp['flowrate'] 
            rcp_str += os.linesep+'flowrate: {}/min'.format(tot_frt)
            solv_frt = tot_frt 
        for itm_nm, val in rcp.items():
            if '_fraction' in itm_nm:
                rg_nm = itm_nm[:itm_nm.find('_fraction')]
                rcp_str += os.linesep+'{} volume fraction: {}'.format(rg_nm,val) 
                rg_frt = tot_frt * val
                solv_frt -= rg_frt
        rcp_str += os.linesep+'{} (solvent) volume fraction: {}'.format(rcp['solvent'],solv_frt/tot_frt) 
        return rcp_str

    @staticmethod
    def print_recipe(rcp):
        rcp_str = ''
        if 'T_set' in rcp:
            rcp_str += 'T_set: {}, '.format(rcp['T_set'])
        if 'T_ramp' in rcp: 
            rcp_str += 'T_ramp: {}/min, '.format(rcp['T_ramp'])
        if 'flowrate' in rcp:
            tot_frt = rcp['flowrate'] 
            rcp_str += 'flowrate: {}/min, '.format(tot_frt)
            solv_frt = tot_frt 
        for itm_nm, val in rcp.items():
            if '_fraction' in itm_nm:
                rg_nm = itm_nm[:itm_nm.find('_fraction')]
                rcp_str += '{} volume fraction: {}'.format(rg_nm,val) 
                rg_frt = tot_frt * val
                solv_frt -= rg_frt
        rcp_str += '{} (solvent) volume fraction: {}'.format(rcp['solvent'],solv_frt/tot_frt) 
        return rcp_str

    def check_status(self):
        ok_flag = True
        stat_str = ''
        stat_dict = {}
        cryo = self.content['cryocon']
        if cryo:
            with cryo.state_lock:
                for chan,idx in cryo['channels'].items():
                    T_set_key = 'T_set_{}'.format(chan)
                    T_read_key = 'T_read_{}'.format(chan)
                    T_set = None
                    if T_set_key in cryo.state:
                        T_set = float(copy.copy(cryo.state[T_set_key]))
                    stat_dict[T_set_key] = T_set
                    T_read = copy.copy(cryo.state[T_read_key])
                    stat_dict[T_read_key] = float(T_read)
                    stat_str += 'T_{}: {} (setpt {}), '.format(chan,T_read,T_set)
        ppcs = self.content['ppumps']
        if any(ppcs):
            for nm, ppc in ppcs.items():
                with ppc.state_lock:
                    setpt_pls = copy.copy(ppc['state']['target_flow_rate'])
                    frt_pls = copy.copy(ppc['state']['flow_rate'])
                if frt_pls is not None:
                    frt_ulm = frt_pls*60/1.E6
                    truefrt_ulm = ppc.get_true_flowrate(frt_ulm)
                    setpt_ulm = setpt_pls*60/1.E6
                    truesetpt_ulm = ppc.get_true_flowrate(setpt_ulm)
                    stat_dict['{}_flowrate'.format(nm)] = float(truefrt_ulm)
                    stat_dict['{}_setpoint'.format(nm)] = float(truesetpt_ulm)
                    stat_str += ' {}: {} (setpt {}), '.format(nm,truefrt_ulm,truesetpt_ulm)
                    if setpt_pls is not None:
                        if setpt_pls > 0:
                            if abs(frt_pls-setpt_pls)/setpt_pls > 0.5 and truesetpt_ulm > 2.:
                                ok_flag = False
                                self.message_callback('ppump {} flowrate {} is far from setpoint {}'
                                .format(nm,truefrt_ulm,truesetpt_ulm))
        tmr = self.content['timer']
        with tmr.dt_lock:
            t_sec = float(tmr.t_utc())
            t_now = float(tmr.dt_utc())
            t_str = str(tmr.time_as_string())
        stat_dict['t_utc'] = float(t_sec)
        stat_dict['time'] = t_str
        #if self.verbose: self.message_callback(stat_str)
        with self.history_lock:
            self.add_to_history(t_now,stat_str)
        return ok_flag,stat_str,stat_dict




import copy
from collections import OrderedDict

from paws.workflows.Workflow import Workflow 

from paws.workflows.IO.BL15 import ReadBatch
from paws.operations.PACKAGING.SortBatch import SortBatch

inputs = copy.deepcopy(ReadBatch.inputs)
inputs.update(
    lower_index = None,
    upper_index = None,
    index_step = 1
    )

outputs = copy.deepcopy(ReadBatch.outputs)

class ReadTimeSeries(Workflow):

    def __init__(self):
        super(ReadTimeSeries,self).__init__(inputs,outputs)
        self.add_operation('read_batch',ReadBatch.ReadBatch())
        self.add_operation('sort',SortBatch())

    def run(self):
        read_inputs = OrderedDict([(k,self.inputs[k]) for k in ReadBatch.inputs.keys()])
        b_outs = self.operations['read_batch'].run_with(**read_inputs) 
        self.operations['sort'].run_with(
            batch_outputs=b_outs,
            x_key='time',
            x_sort_flag=True,
            x_shift_flag=True,
            lower_index=self.inputs['lower_index'],
            upper_index=self.inputs['upper_index'],
            index_step=self.inputs['index_step']
            )
        self.outputs.update(self.operations['sort'].outputs['sorted_outputs'])
        return self.outputs




from __future__ import print_function
from collections import OrderedDict
import sys
import os
import copy
from threading import Thread, Condition
if int(sys.version[0]) == 2:
    import Queue as queue
else:
    import queue 

import numpy as np
import tzlocal
import datetime

from .. import pawstools

# TODO: clones pass history back to proxy,
# proxy keeps the log file.
# log files are created at start().

# TODO: ensure plugins run clean either as clones or standalone

class PawsPlugin(object):
    """Base class for building PAWS Plugins."""

    def __init__(self,content):
        super(PawsPlugin,self).__init__()
        self.content = OrderedDict(copy.deepcopy(content))
        self.content_doc = OrderedDict.fromkeys(self.content.keys()) 
        self.message_callback = self.tagged_print
        self.data_callback = None 
        # take the running_lock when toggling the run flag 
        self.running_lock = Condition()
        self.running = False
        # use command_lock to add or pop commands from the queue
        self.command_lock = Condition()
        self.commands = queue.Queue() 
        # use history lock to edit the history
        self.history_lock = Condition()
        self.history = []
        self.n_events = 0
        # proxy is the thread_clone's ref to this object, 
        # i.e. self.thread_clone.proxy is self.
        self.thread_clone = None
        self.proxy = None
        # keep track of py version for convenience
        self.py_version = int(sys.version[0])
        self.log_file = None
        self.verbose = False
        self.thread_blocking = False
        self.tz = tzlocal.get_localzone()
        self.ep = datetime.datetime.fromtimestamp(0,self.tz)
        t0 = datetime.datetime.now(self.tz)
        self.t0_utc = (t0-self.ep).total_seconds()

    def t_utc(self):
        t_now = datetime.datetime.now(self.tz)
        t_utc = (t_now-self.ep).total_seconds()
        return t_utc

    def dt_utc(self):
        t_utc_now = self.t_utc()
        return t_utc_now-self.t0_utc

    def get_date_time(self):
        return str(datetime.datetime.now(self.tz))

    def set_log_file(self,file_path=None):
        log_file = '{}_{}.log'.format(type(self).__name__,int(self.t0_utc*1000))
        if not file_path:
            file_path = os.path.join(pawstools.paws_scratch_dir,log_file)
        if file_path == self.log_file:
            return
        suffix = 1
        pth,ext = os.path.splitext(file_path)
        while os.path.exists(file_path):
            file_path = pth+'_{}'.format(suffix)+ext
            suffix += 1
        open(file_path,'a').close()
        self.message_callback('plugin log file: {}'.format(file_path))
        self.log_file = file_path 

    def set_log_dir(self,dir_path): 
        dp,fn = os.path.split(self.log_file)
        self.set_log_file(os.path.join(dir_path,fn))

    #def set_log_file(self,file_name):
    #    dp,fn = os.path.split(self.log_file)
    #    self.set_log_file(os.path.join(dp,file_name))

    def set_verbose(self,verbose_flag):
        if verbose_flag:
            self.verbose = True
        else:
            self.verbose = False 

    def set_data(self,item_key,val):
        key_parts = item_key.split('.')
        itm = self.content
        for p in key_parts[:-1]:
            itm = itm[p]
        itm[key_parts[-1]] = val

    def __getitem__(self,key):
        return self.content[key]
    def keys(self):
        return self.content.keys() 
    def __setitem__(self,key,val):
        self.set_data(key,val)

    def tagged_print(self,msg):
        print('[{}] {}'.format(type(self).__name__,msg))

    def description(self):
        """Describe the plugin.

        PawsPlugin.description() returns a string 
        documenting the functionality of the PawsPlugin,
        the current input settings, etc.
        Reimplement this in PawsPlugin subclasses.
        """
        return str(self.setup_dict()) 

    def start(self):
        """Start the plugin.

        Assuming a plugin's content has been properly set,
        PawsPlugin.start() should prepare the plugin for use, 
        e.g. by opening connections, reading files, etc. 
        Reimplement this in PawsPlugin subclasses as needed.
        It can be useful to call super().start() 
        for a default log file and proper thread setup.
        """
        self.set_log_file()
        self.running = True 
        if self.thread_blocking: 
            self.run_clone()
        else:
            self.run()

    def run(self):
        """Use the plugin in a continuing thread.
    
        Plugins with some continuous function should implement run().
        During start(), plugins that must run continually may be cloned, 
        and the continuous functionality is carried out by clone.run(), 
        with thread locks used to share data with the original plugin.
        """
        pass

    def run_clone(self):
        with self.running_lock:
            self.running = True
        #self.thread_clone = self.build_clone()
        #self.thread_clone.proxy = self
        self.thread_clone.log_file = self.log_file
        th = Thread(target=self.thread_clone.run)
        th.start()

    def run_notify(self):
        with self.running_lock:
            if int(self.py_version) == 2:
                self.running_lock.notifyAll()
            else:
                self.running_lock.notify_all()

    def stop(self):
        """Stop the plugin.

        PawsPlugin.stop() should provide a clean end for the plugin,
        for instance closing connections used by the plugin.
        Reimplement this in PawsPlugin subclasses as needed.
        """
        with self.running_lock:
            self.running = False 
        if self.thread_clone:
            self.thread_clone.stop()

    @classmethod
    def clone(cls):
        return cls()

    def build_clone(self):
        """Clone the Plugin."""
        new_pgn = self.clone()
        for nm,val in self.content.items():
            new_pgn.content[nm] = copy.deepcopy(val) 
        new_pgn.verbose = bool(self.verbose)
        #new_pgn.data_callback = self.data_callback
        #new_pgn.message_callback = self.message_callback
        return new_pgn

    # TODO: refactor: optional timestamp, if None get current utc
    def add_to_history(self,t,msg):
        """Add a timestamp and a message to the plugin's history."""
        self.history.append((t,msg))
        self.n_events += 1
        if np.mod(self.n_events,100) == 0:
            self.dump_history(90)
            # always keep the past 10 points?
            self.history = self.history[-10:]

    def dump_history(self,n_events=None):
        dump_file = open(self.log_file,'a')
        h = self.history
        if n_events is not None: h = self.history[:n_events]
        for t,msg in h:
            dump_file.write('{}: {}'.format(t,msg)+os.linesep)
        dump_file.close()

    def notify_locks(self,locks):
        if not isinstance(locks,list):
            locks = [locks]
        for l in locks:
            with l:
                if int(self.py_version) == 2:
                    l.notifyAll()
                else:
                    l.notify_all()




import os
from collections import OrderedDict
from threading import Condition

import numpy as np
import pyFAI

from .PawsPlugin import PawsPlugin

content = OrderedDict(calib_file=None) 

class PyFAIIntegrator(PawsPlugin):
    """Plugin for applying a PyFAI.AzimuthalIntegrator.

    Input calibration file should be in one of the formats
    outlined in the package documentation. 
    """
    def __init__(self):
        super(PyFAIIntegrator,self).__init__(content)
        self.content_doc['calib_file'] = 'file defining a dict of calibration parameters, '\
            'in one of the formats outlined in the package documentation'
        self.integrator_lock = Condition()
        self.integrator = None

    def start(self):
        super(PyFAIIntegrator,self).start()
        with self.integrator_lock:
            self.integrator = pyFAI.AzimuthalIntegrator()
        self.set_calib()

    def set_calib_file(self,file_path):
        self.content['calib_file'] = file_path
        #self.set_calib()

    def set_calib(self):
        self.message_callback('calibrating on {}'.format(self.content['calib_file']))
        calib = self.content['calib_file']
        fp,xt = os.path.splitext(calib)
        #print(xt)
        if xt in ['.poni','.PONI']:
            #g = pyFAI.geometry.Geometry()
            #g.read(calib)
            #p.setPyFAI(g.getPyFAI())
            with self.integrator_lock:
                self.integrator.read(calib)
        elif xt in ['.nika','.NIKA']:
            with self.integrator_lock:
                self.set_nika(calib)

    def integrate_to_1d(self,img_data,npt=1000,polz_factor=0.,unit='q_A^-1'):
        with self.integrator_lock:
            self.message_callback('integrating image...')
            q,I = self.integrator.integrate1d(img_data,npt,
                polarization_factor=polz_factor,unit=unit)
        return q,I

    def integrate_to_2d(self,img_data,npt_rad=1000,npt_azim=1000,polz_factor=0.,unit='q_A^-1'):
        with self.integrator_lock:
            I_at_q_chi,q,chi = self.integrator.integrate2d(img_data,
                npt_rad,npt_azim,polarization_factor=polz_factor,unit=unit)
        return q,chi,I_at_q_chi

    def set_nika(self,nika_file):
        # TODO: make nika format yaml-able
        for line in open(nika_file,'r'):
            kv = line.strip().split('=')
            if 'sample_to_CCD_mm' in kv[0]:
                d_mm = float(kv[1])
            if 'pixel_size_x_mm' in kv[0]:
                pxsz_x_mm = float(kv[1])
            if 'pixel_size_y_mm' in kv[0]:
                pxsz_y_mm = float(kv[1])
            if 'beam_center_x_pix' in kv[0]:
                bcx_px = float(kv[1])       # Nika reports the x coord relative to 'bottom left' corner of detector
                                            # where beam axis intersects detector plane, in pixels 
            if 'beam_center_y_pix' in kv[0]:
                bcy_px = float(kv[1])       # same as beam_center_x_pix but for y 
            if 'horizontal_tilt_deg' in kv[0]:    
                htilt_deg = float(kv[1])    # Nika reports the horizontal tilt in degrees...
            if 'vertical_tilt_deg' in kv[0]:
                vtilt_deg = float(kv[1])    # Nika reports the vertical tilt in degrees...
            if 'wavelength_A' in kv[0]:
                wl_A = float(kv[1])         # Nika reports wavelength is in Angstroms
        wl_m = wl_A*1E-10
        pxsz_x_um = pxsz_x_mm * 1000
        pxsz_y_um = pxsz_y_mm * 1000
        # TODO: check whether these rotation angle mappings are correct,
        # going from nika to fit2D geometry. 
        tilt_deg = -1.*htilt_deg
        rot_fit2d = vtilt_deg
        #tmpint = pyFAI.AzimuthalIntegrator(wavelength = wl_m)
        #tmpint.setFit2D(d_mm,bcx_px,bcy_px,tilt_deg,rot_fit2d,pxsz_x_um,pxsz_y_um)
        #pd = tmpint.getPyFAI()
        #self.integrator.setPyFAI(**pd)
        self.integrator.set_wavelength(wl_m)
        self.integrator.setFit2D(d_mm,bcx_px,bcy_px,tilt_deg,rot_fit2d,pxsz_x_um,pxsz_y_um)

    def get_plugin_content(self):
        d = super(PyFAIIntegrator,self).get_plugin_content()
        d['integrator'] = self.integrator 
        return d





from __future__ import print_function
import numpy as np
from collections import OrderedDict

import pyFAI

from ...Operation import Operation

inputs = OrderedDict(
    image_data=None,
    integrator=None,
    npt=1000,
    polz_factor=0.,
    unit='q_A^-1')
outputs = OrderedDict(
    q=None,
    I=None,
    q_I=None)
        
class Integrate1d(Operation):
    """Integrate an image using a PyFAIIntegrator plugin.
    
    Input image data (ndarray), PyFAI.AzimuthalIntegrator, 
    and other parameters for PyFAI.AzimuthalIntegrator.integrate1d().
    
    Refer to the PyFAI documentation at 
    http://pyfai.readthedocs.io/en/latest/ 
    for supported keyword arguments 
    as well as parameter definitions and defaults.
    """

    def __init__(self):
        super(Integrate1d,self).__init__(inputs,outputs)
        self.input_doc['image_data'] = '2d array representing intensity for each pixel'
        self.input_doc['integrator'] = 'A PyFAIIntegrator PawsPlugin'
        self.input_doc['npt'] = 'number of q-points to integrate'
        self.input_doc['polz_factor'] = 'polarization factor'
        self.input_doc['unit'] = 'choice of unit: see PyFAI documentation for options' 
        self.output_doc['q'] = 'Scattering vector magnitude q in 1/Angstrom'
        self.output_doc['I'] = 'Integrated intensity at q'
        self.output_doc['q_I'] = 'q and I together as n-by-2 numpy array'

    def run(self):
        img = self.inputs['image_data']
        npt = self.inputs['npt']
        intgtr = self.inputs['integrator']
        pzfac = self.inputs['polz_factor']
        u = self.inputs['unit']
        q,I = intgtr.integrate_to_1d(img,npt,polz_factor=pzfac,unit=u)
        self.outputs['q'] = q 
        self.outputs['I'] = I
        self.outputs['q_I'] = np.array([q,I]).T




from collections import OrderedDict
import copy
import os

from xrsdkit.tools import ymltools as xrsdyml

from ...Workflow import Workflow 
from ....operations.IO.YAML.LoadYAML import LoadYAML
from ....operations.IO.IMAGE.FabIOOpen import FabIOOpen
from ....operations.IO.NumpyLoad import NumpyLoad

# NOTE: this workflow is for reading samples
# that were saved with YAML headers

inputs = OrderedDict(
    header_file = None,
    time_key = 't_utc',
    image_file = None,
    q_I_file = None,
    system_file = None
    )

outputs = OrderedDict(
    time = None,
    header_data = None,
    image_data = None,
    q_I = None,
    dI = None,
    system = None
    )

class Read(Workflow):

    def __init__(self):
        super(Read,self).__init__(inputs,outputs)
        self.add_operations(
            read_header=LoadYAML(),
            read_image=FabIOOpen(),
            read_q_I=NumpyLoad(),
            )

    def run(self):
        self.outputs = copy.deepcopy(outputs)

        if self.ops_enabled['read_header']:
            if (self.inputs['header_file']) and (os.path.exists(self.inputs['header_file'])):
                hdr_outs = self.operations['read_header'].run_with(
                file_path=self.inputs['header_file'])
                hdata = hdr_outs['data']
                self.outputs['header_data'] = hdata
                if hdata and self.inputs['time_key']:
                    self.outputs['time'] = hdata[self.inputs['time_key']] 
            else:
                self.message_callback('header file not found: {}'.format(self.inputs['header_file']))

        if self.ops_enabled['read_image']:
            if (self.inputs['image_file']) and (os.path.exists(self.inputs['image_file'])):
                img_outs = self.operations['read_image'].run_with(
                file_path=self.inputs['image_file'])
                self.outputs['image_data'] = img_outs['image_data']
            else:
                self.message_callback('image file not found: {}'.format(self.inputs['image_file']))

        if self.ops_enabled['read_q_I']:
            if (self.inputs['q_I_file']) and (os.path.exists(self.inputs['q_I_file'])):
                q_I_outs = self.operations['read_q_I'].run_with(
                file_path=self.inputs['q_I_file'])
                q_I = q_I_outs['data']
                dI = None
                if (q_I is not None) and (q_I.shape[1] > 2):
                    q_I = q_I[:,:2]
                    dI = q_I[:,2]
                self.outputs['q_I'] = q_I
                self.outputs['dI'] = dI
            else:
                self.message_callback('q_I file not found: {}'.format(self.inputs['q_I_file']))

        if (self.inputs['system_file']) and (os.path.exists(self.inputs['system_file'])):
            self.outputs['system'] = xrsdyml.load_sys_from_yaml(self.inputs['system_file'])
        else:
            self.message_callback('xrsd system file not found: {}'.format(self.inputs['system_file']))

        return self.outputs




from __future__ import print_function
from collections import OrderedDict
import os
import datetime
import time

import tzlocal

from ...Operation import Operation

inputs=OrderedDict(
    file_path=None,
    temperature_key='TEMP')

outputs=OrderedDict(
    data=None,
    dir_path=None,
    filename=None)

class ReadSpecHeader(Operation):
    """
    Read a .txt header from beamline 1-5 at SSRL into a dict.
    """

    def __init__(self):
        super(ReadSpecHeader, self).__init__(inputs, outputs)
        self.input_doc['file_path'] = 'path to a .txt header file produced by beamline 1-5 at SSRL.'
        self.output_doc['data'] = 'the header data, packaged as a python dictionary'
        self.output_doc['dir_path'] = 'directory path'
        self.output_doc['filename'] = 'filename with path and extension stripped'

    def run(self):
        p = self.inputs['file_path']
        dir_path,filename = os.path.split(p)
        file_noext = os.path.splitext(filename)[0]
        self.outputs['filename'] = file_noext 
        self.outputs['dir_path'] = dir_path 
        d = OrderedDict()
        self.message_callback('reading {}'.format(p))
        for l in open(p,'r').readlines():
           if not l.strip() == '' and not l.strip()[0] == '#':
                kvs = l.split(',')
                # special case for the string headers on line 1
                if not kvs[0].find('User') == -1:
                    u_str = kvs[0].split('User:')[1].strip()
                    t_str = kvs[1].split('time:')[1].strip()
                    d['User'] = u_str
                    d['date_time'] = t_str
                    tz = tzlocal.get_localzone()
                    # use strptime to create a naive datetime
                    dt = datetime.datetime.strptime(t_str.strip(),"%a %b %d %H:%M:%S %Y")
                    # add timezone information to datetime
                    dt_aware = datetime.datetime(dt.year,dt.month,dt.day,dt.hour,dt.minute,dt.second,dt.microsecond,tz)
                    # interpret the time in UTC 
                    t_utc = time.mktime(dt_aware.timetuple())
                    d['t_utc'] = float(t_utc)
                # and filter out the redundant temperature line
                elif not (len(kvs)==1 and kvs[0].strip()[-1]=='C'):
                    for kv in kvs:
                        kv_arr = kv.split('=')
                        d[kv_arr[0].strip()] = float(kv_arr[1].strip())
        # add 'temperature' to the output data,
        # to make a consistent interface with the non-legacy header reader
        if self.inputs['temperature_key'] in d:
            d['temperature'] = d[self.inputs['temperature_key']]
        else:
            d['temperature'] = None
        self.outputs['data'] = d
        return self.outputs




from collections import OrderedDict
import os
import time

from ...Operation import Operation

inputs=OrderedDict(
    ppump_controllers={},
    targets={},
    set_points={})
outputs=OrderedDict(
    report=None)
        
class SetPPumps(Operation):
    """Set the flow or pressure for an array of P-pump controllers."""

    def __init__(self):
        super(SetPPumps,self).__init__(inputs,outputs)
        self.input_doc['ppump_controllers'] = 'dict of MitosPPumpController plugins'
        self.input_doc['targets'] = 'dict of control modes (strings) for each pump- '\
            'each entry should be either "flowrate" or "pressure", '\
            'defaults to "flowrate" if values are not provided' 
        self.input_doc['set_points'] = 'dict of set points for each pump' 
        #self.input_doc['delay_time'] = 'seconds to wait after setting `set_points`' 
        self.output_doc['report'] = 'dict reporting details of final state' 

    def run(self):
        ppcs = self.inputs['ppump_controllers'] 
        tgts = self.inputs['targets']
        setpts = self.inputs['set_points']
        #delay = self.inputs['delay_time']
        #stat = self.inputs['flag']
        #vals = [None for ppc in ppcs] 
        #if bool(stat):
        if tgts is None:
            tgts = ['flowrate' for ppc in ppcs]
        self.message_callback('setting pumps to: {}'.format(setpts))
        for ppnm,ppc in ppcs.items():
            tgt = tgts[ppnm]
            setpt = setpts[ppnm]
            if tgt == 'pressure':
                ppc.set_pressure(setpt)
            #if tgt == 'flowrate':
            else: 
                ppc.set_flowrate(setpt)
        #self.message_callback('SetPPumps waiting {} seconds...'.format(delay))
        #time.sleep(delay)
            #done = False
            #while not done:
            #    done = True
            #    for ipp,ppc in enumerate(ppcs):
            #        tgt = tgts[ipp]
            #        setpt = setpts[ipp]
            #        prec = precs[ipp]
            #        vals[ipp] = ppc.get_current_value(tgt)
            #        if not abs(vals[ipp]-setpt)/setpt < prec:
            #            done = False
            #    # TODO:
            #    # delay, then check again
            #    # implement a maximum wait time
            #rpt = dict(targets=tgts,final_values=vals)
        #    rpt = {} 
        #    self.outputs['report'] = rpt
        #    self.outputs['flag'] = True
        #else:
        #    self.outputs['report'] = {'STATUS':bool(stat)}
        #    self.outputs['flag'] = False 




import os

from paws.workflows.INTEGRATION.Integrate import Integrate 
from paws.plugins.PluginManager import PluginManager
from paws import pawstools

tdd = os.path.join(pawstools.rootdir,'tests','test_data')

def test():
    pgmgr = PluginManager()
    pgmgr.add_plugin('integrator','PyFAIIntegrator')
    wf = Integrate()
    outs = wf.run_with(
        image_file = os.path.join(tdd,'images','test1.tif'),
        integrator = pgmgr.plugins['integrator'],
        n_points = 30,
        polz_factor = 0.9,
        calib_file = os.path.join(tdd,'calib','test.nika'),
        output_dir = os.path.join(tdd,'outputs')
        )
    return True




from collections import OrderedDict
import warnings

from fabio import edfimage
from fabio import tifimage

from ...Operation import Operation
import os

inputs=OrderedDict(
    image_data=None,
    header=None,
    dir_path=None,
    filename=None,
    filetag='',
    ext='.tif',
    overwrite=False)
outputs = OrderedDict(file_path=None)

class FabIOWrite(Operation):
    """
    Use FabIO to write out an image, given image data,
    directory path, filename, file tag, extension, 
    an image header (dict), and a flag for whether or not to overwrite. 

    Outputs the full file path where the image was written,
    which should be dir_path+filename+filetag+ext.
    """

    def __init__(self):
        super(FabIOWrite,self).__init__(inputs,outputs)
        self.input_doc['image_data'] = 'image/array data to be saved'
        self.input_doc['header'] = 'dict-like image metadata header'
        self.input_doc['dir_path'] = 'parent directory path'
        self.input_doc['filename'] = 'base filename to be saved as'
        self.input_doc['filetag'] = 'string tag to be appended to base filename'
        self.input_doc['ext'] = 'file extension (overwrites base filename extension)'
        self.input_doc['overwrite'] = 'allow overwrite of already existing files'
        self.output_doc['file_path'] = 'path to the file that will be written: dir_path+filename+filetag+ext'

    def run(self):
        """
        Call on fabIO to extract image data
        """
        img = self.inputs['image_data']
        p = self.inputs['dir_path']
        fnm = self.inputs['filename'] 
        ext = self.inputs['ext'].lower()
        tg = self.inputs['filetag']
        hdr = self.inputs['header']
        outfile = fnm + tg + ext
        filepath = os.path.join(self.inputs['dir_path'],outfile)
        self.outputs['file_path'] = filepath
        if os.path.isfile(filepath) and not self.inputs['overwrite']:
            warnings.warn('Skipping output for {}: '\
            'File exists and overwrite is disabled'.format(filepath))
        cls = None
        if ext == '.edf': cls = edfimage.edfimage
        elif ext == '.tif': cls = tifimage.tifimage

        if not cls: raise ValueError('Extension not supported.')

        cls(img,header=hdr).write(filepath)





import numpy as np
from collections import OrderedDict

from ...Operation import Operation

inputs = OrderedDict(
    q_I=None,
    q_I_bg=None,
    dI=None,
    dI_bg=None)
outputs = OrderedDict(
    q_I_bgsub=None,
    dI=None,
    bg_factor=None)

class BgSubtract(Operation):
    """Background subtraction for 1-d spectra (intensity versus q).

    Subtracts the background from an image,
    after scaling the background to prevent over-subtraction.
    Given error estimates for the image and background, 
    estimates the error for the background-subtracted intensity. 
    """

    def __init__(self):
        super(BgSubtract, self).__init__(inputs, outputs)
        self.input_doc['q_I'] = 'n-by-2 array of q values and corresponding intensity values'
        self.input_doc['q_I_bg'] = 'n-by-2 array, background corresponding to q_I'
        self.input_doc['dI'] = '1d array, error estimate of I (optional, default None)' 
        self.input_doc['dI_bg'] = '1d array, error estimate of I_bg (optional, default None)'
        self.output_doc['q_I_bgsub'] = 'n-by-2 array of q values and background-subtracted intensity: I-(bg_factor*I_bg)'
        self.output_doc['dI'] = 'error estimate of background-subtracted intensity'
        self.output_doc['bg_factor'] = 'the factor the background was multiplied by '\
            'before subraction to ensure positive values for output intensity'

    def run(self):
        q_I = self.inputs['q_I']
        q_I_bg = self.inputs['q_I_bg']
        I = q_I[:,1]
        I_bg = q_I_bg[:,1]
        bad_data = (I < 0) | (I_bg <= 0) | np.isnan(I) | np.isnan(I_bg)
        bg_factor = np.min(I[~bad_data] / I_bg[~bad_data])
        #if bg_factor > 1.: bg_factor = 1.
        self.message_callback('subtracting background (bg multiplier: {})'.format(bg_factor))
        I_out = I-(bg_factor*I_bg)
        dI = self.inputs['dI']
        dI_bg = self.inputs['dI_bg']
        dI_out = None
        if dI_bg is not None and dI is not None:
            dI_out = (dI**2+(bg_factor*dI_bg)**2)**0.5
        q_I_bgsub = np.zeros(q_I.shape)
        q_I_bgsub[:,0] = q_I[:,0]
        q_I_bgsub[:,1] = I_out
        self.outputs['q_I_bgsub'] = q_I_bgsub 
        self.outputs['dI'] = dI_out
        self.outputs['bg_factor'] = bg_factor
        return self.outputs




import os

from paws.workflows.IO.BL15.Read import Read
from paws import pawstools

tdd = os.path.join(pawstools.rootdir,'tests','test_data')

def test():
    read_wf = Read()
    outs = read_wf.run_with(
        time_key = 't_utc',
        header_file = os.path.join(tdd,'headers','test1.yml'),
        image_file = os.path.join(tdd,'images','test1.tif'),
        q_I_file = os.path.join(tdd,'data','test1.dat'),
        system_file = os.path.join(tdd,'systems','test1.yml')
        )
    return True




from __future__ import print_function
from collections import OrderedDict
from functools import partial
import importlib

from ..pawstools import DictTree
from .. import plugins as pgns

class PluginManager(DictTree):
    """Tree for storing, browsing, and managing PawsPlugins"""

    def __init__(self):
        super(PluginManager,self).__init__()
        self.plugins = self._root
        self.plugins_running = OrderedDict()
        self.connections = OrderedDict()
        self.message_callback = self.tagged_print 

    def tagged_print(self,msg):
        print('[{}] {}'.format(type(self).__name__,msg))

    def set_verbose(self,plugin_name,flag=True):
        self.plugins[plugin_name].set_verbose(flag)

    def set_log_dir(self,dir_path):
        for nm,pg in self.plugins.items():
            pg.set_log_dir(dir_path)

    def add_plugin(self,plugin_name,plugin_module):
        """Import, name, and add a plugin.

        After a plugin is added to a plugin_manager,
        it is available as plugin_manager.plugins[plugin_name].

        Parameters
        ----------
        plugin_name : str
            Name that will be used to refer to this plugin after it is added.
        plugin_module : str
            Name of the plugin module.
            Example: If class MyPlugin is in the CATEGORY.MyPlugin module,
            retrieve it with `plugin_module` = 'CATEGORY.MyPlugin'.
        """
        p = self.get_plugin(plugin_module)
        if not self.is_key_valid(plugin_name): 
            raise KeyError(self.key_error_message(plugin_name))
        #p.message_callback = self.message_callback
        p.data_callback = partial(self.set_plugin_item,plugin_name)
        self.set_data(plugin_name,p)
        self.plugins_running[plugin_name] = False 
    
    def add_plugins(self,**kwargs):
        for pgn_nm,pgn_mod in kwargs.items():
            self.add_plugin(pgn_nm,pgn_mod)

    def get_plugin(self,plugin_module): 
        """Import, instantiate, return a PawsPlugin from its module.

        This can also be used to test the Python environment 
        for compatibility with a plugin.

        Parameters
        ----------
        plugin_module : str
            Name of the plugin module.
            See add_plugin().

        Returns
        -------
        PawsPlugin 
            An instance of the PawsPlugin subclass defined in `plugin_module`. 
        """
        mod = importlib.import_module('.'+plugin_module,pgns.__name__)
        return mod.__dict__[plugin_module]()

    def set_plugin_contents(self,plugin_name,**kwargs):
        for nm,val in kwargs.items():
            self.set_plugin_item(plugin_name,nm,val)

    #def set_plugin_content(self,plugin_name,item_name,val):
    #    """Set an item of plugin content to the provided value.
    #
    #    Parameters
    #    ----------
    #    plugin_name : str
    #        Name that will be used to refer to this plugin after it is added.
    #    item_name : str
    #        name of the plugin content item to be set
    #    val : object
    #        the data to be referenced to the specified item 
    #    """
    #    self.set_data(plugin_name+'.'+item_name,val)

    def set_plugin_item(self,pgn_name,item_key,item_data):
        full_key = pgn_name+'.'+item_key
        self.set_data(full_key,item_data)

    def connect(self,item_key,content_map):
        """Connect the data at `item_key` to one or more plugin content items.

        Sets up Plugin content listed in `content_map`
        to take the value at `item_key`.
        `content_map` can be a DictTree key (string) or a list thereof.
        """
        if item_key in self.connections:
            if isinstance(content_map,list):
                self.connections[item_key].extend(content_map)
            else:
                self.connections[item_key].append(content_map)
        else:
            if not isinstance(content_map,list): content_map = [content_map]
            self.connections[item_key] = content_map

    def start_plugins(self,plugin_name_list):
        for pn in plugin_name_list:
            self.start_plugin(pn)

    def start_plugin(self,plugin_name):
        """Start the plugin referred to by `plugin_name`."""
        self.message_callback('starting plugin {}'.format(plugin_name))
        pgn = self.prepare_plugin(plugin_name)
        self.plugins_running[plugin_name] = True 
        pgn.start()
        
    def prepare_plugin(self,plugin_name):
        pgn = self.plugins[plugin_name]
        if pgn.thread_blocking:
            pgn_clone = self.plugins[plugin_name].build_clone()
            pgn.thread_clone = pgn_clone 
            pgn.thread_clone.proxy = pgn
        for item_key,content_map in self.connections.items():
            for content_key in content_map:
                if content_key.split('.')[0] == plugin_name:
                    self.set_data(content_key,self.get_data(item_key))
                    if pgn.thread_blocking:
                        pgn_item_key = content_key[content_key.find(plugin_name)+len(plugin_name)+1:]
                        pgn_clone.set_data(pgn_item_key,self.get_data(item_key))
        return pgn

    def stop_plugin(self,plugin_name):
        self.message_callback('stopping plugin {}'.format(plugin_name))
        self.plugins[plugin_name].stop()
        self.plugins_running[plugin_name] = False 

    def setup_dict(self):
        pg_dict = OrderedDict()
        for pg_name,pg in self.plugins.items():
            pg_dict[pg_name] = OrderedDict()
            pg_dict[pg_name]['MODULE'] = pg.__module__[pg.__module__.find('plugins.')+8:] 
            pg_dict[pg_name]['CONTENT'] = pg.content
        pg_dict['CONNECTIONS'] = self.connections 
        return pg_dict

    def load_plugins(self,pg_dict):
        """Load plugins from a setup dict.

        Written for loading plugins from saved data.

        Parameters
        ----------
        pg_dict : dict 
            Dict specifying plugin setup
        """
        for item_key,content_map in pg_dict.pop('CONNECTIONS').items():
            self.connect(item_key,content_map)
        for pg_name in pg_dict.keys():
            self.add_plugin(pg_name,pg_dict[pg_name]['MODULE'])
            for item_name, val in pg_dict[pg_name]['CONTENT'].items():
                self.set_plugin_content(pg_name,item_name,val)

    def n_plugins(self):
        """Return number of plugins currently loaded."""
        return len(self.plugins) 





from collections import OrderedDict

from ...Operation import Operation

from xrsdkit.tools import piftools
from xrsdkit.system import System

inputs=OrderedDict(
    system=None,
    q_I_file=None,
    q_I=None,
    #experiment_id=None,
    t_utc=None,
    temperature=None,
    source_wavelength=None,
    )
outputs=OrderedDict(pif=None)

class PackXRSDPIF(Operation):
    """Build a PIF record for an xrsdkit System and its I(q) pattern"""

    def __init__(self):
        super(PackXRSDPIF,self).__init__(inputs,outputs)
        self.input_doc['system'] = 'xrsdkit.system.System object describing scattering material, '\
            'or a dict containing the System data, in which case the dict is used '\
            'to build the system by calling the xrsdkit.system.System constructor'
        self.input_doc['q_I_file'] = 'path to q_I data file (within the Citrination dataset)'
        self.input_doc['q_I'] = 'n-by-2 array of scattering data, intensity (arb) versus q (1/A)'
        #self.input_doc['experiment_id'] = 'string experiment id, used as a pif ID attribute'
        self.input_doc['t_utc'] = 'time in seconds utc'
        self.input_doc['temperature'] = 'temperature of the system in degrees Celsius'
        self.input_doc['source_wavelength'] = 'light source wavelength in Angstroms'
        self.output_doc['pif'] = 'pif object representing the input data'

    def run(self):
        sys = self.inputs['system']
        if isinstance(sys,dict):
            sys = System(sys)
        q_I = self.inputs['q_I']
        t_utc = self.inputs['t_utc']
        temp_C = self.inputs['temperature']
        src_wl = self.inputs['source_wavelength']
        expt_id = 'tmp'
        if sys.sample_metadata['experiment_id']:
            expt_id = sys.sample_metadata['experiment_id']
        sample_id = 'tmp'
        if sys.sample_metadata['sample_id']:
            sample_id = sys.sample_metadata['sample_id']
        csys = piftools.make_pif(sample_id,sys,self.inputs['q_I_file'],q_I,sys.sample_metadata['experiment_id'],t_utc,temp_C,src_wl)
        self.outputs['pif'] = csys
        return self.outputs




"""
paws setup module
---- ----- ------
from https://packaging.python.org/distributing/
"""

#from setuptools import setup, find_packages
# To use a consistent encoding
#from codecs import open

from os import path
from setuptools import setup, find_packages

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here,'README.rst')) as f:
    long_description = f.read()

# Get authorship as string from contributors.txt
__authorship__=''
with open(path.join(here,'contributors.txt')) as f: 
    for line in f.readlines():
        __authorship__ += line.strip()+', '
__authorship__ = __authorship__[:-2]

# Executing paws/config.py defines __version__ 
with open(path.join(here,'paws','config.py')) as f: 
    exec(f.read())

setup(
    name='pypaws',
    version=__version__,
    description='the Platform for Automated Workflows by SSRL',
    long_description=long_description,
    url='https://github.com/slaclab/paws/',
    author=__authorship__,
    author_email='paws-developers@slac.stanford.edu',
    license='BSD',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',

        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',
        
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],

    keywords='data analysis workflow',

    packages=find_packages(exclude=[]),
    #py_modules=["my_module"],

    package_dir={'paws':'paws'},

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[\
        'pyyaml','numpy','scipy',\
        'fabio','pyfai','xrsdkit',\
        'tzlocal','pyserial','serial','paramiko',\
        'pypif','citrination_client'\
        ],
    python_requires='>=2.7',

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    #extras_require={
    #    'dev': ['check-manifest'],
    #    'test': ['coverage'],
    #},

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    include_package_data = True,
    package_data={
        '': ['contributors.txt'],
        #'': ['contributors.txt','all_requirements.txt'],
        #'': ['all_requirements.txt'],
        #'paws': ['qt/graphics/*.png','qt/qtui/*.ui'],
    },

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. 
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    #data_files=[('', ['contributors.txt'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    #entry_points={
    #    'console_scripts': [
    #        #'paws=main:main',
    #        # TODO: add a jupyter/ipython console entry point 
    #        #'pawsconsole=paws.qt.widgets.widget_launcher:ipypaws',
    #    ],
    #},
)



import numpy as np
from collections import OrderedDict

from ...Operation import Operation
from xrsdkit.tools import g_of_r

inputs = OrderedDict(q_I=None)
outputs = OrderedDict(
    r_gofr=None,
    r_max=None)

class FFTRadialDistribution(Operation):
    """Obtain the radial distribution function from a SAXS pattern via FFT."""

    def __init__(self):
        super(FFTRadialDistribution, self).__init__(inputs, outputs)
        self.input_doc['q_I'] = 'n-by-2 array of q (1/Angstrom) versus I.'
        self.output_doc['r_gofr'] = 'n-by-2 array of r (Angstrom) versus g(r).'
        self.output_doc['r_max'] = 'the integral from zero to r_max of g(r) '\
                            'is 0.99 times the full integral of g(r)'

    def run(self):
        q_I = self.inputs['q_I']
        r_gofr,r_max = saxs_math.g_of_r(q_I)
        self.outputs['r_gofr'] = r_gofr
        self.outputs['r_max'] = r_max




from .. import TimeSeriesXRSDFit as TimeSeriesXRSDFit_new
from ....IO.BL15.LEGACY.ReadTimeSeries import ReadTimeSeries 

class TimeSeriesXRSDFit(TimeSeriesXRSDFit_new.TimeSeriesXRSDFit):

    def __init__(self):
        super(TimeSeriesXRSDFit,self).__init__()
        self.add_operation('read_timeseries',ReadTimeSeries())




import os

from paws import pawstools
from paws.workflows.WfManager import WfManager

wfmgr = WfManager()

#wfmgr.load_packaged_workflow('read','IO.BL15.read')
wfmgr.load_packaged_wfm('FITTING.BL15.batch_fit')
# replace YAML-based read workflow with legacy version 
wfmgr.load_packaged_workflow('read','IO.BL15.LEGACY.read')

wfmgr.save_to_wfm(os.path.join(pawstools.sourcedir,'workflows','FITTING','BL15','LEGACY','batch_fit.wfm'))




from collections import OrderedDict
import os

import fabio

from ...Operation import Operation

inputs=OrderedDict(file_path=None)
outputs=OrderedDict(
    image_data=None,
    FabioImage=None,
    dir_path=None,
    filename=None)

class FabIOOpen(Operation):
    """
    Takes a filesystem path and calls fabIO to load it. 
    """

    def __init__(self):
        super(FabIOOpen,self).__init__(inputs,outputs) 
        self.input_doc['file_path'] = 'string representing the path to a .tif image'
        self.output_doc['image_data'] = '2D array representing pixel values taken from the input file'
        self.output_doc['FabioImage'] = 'The object generated by fabio.open()'
        self.output_doc['dir_path'] = 'Path to the directory the image came from'
        self.output_doc['filename'] = 'The image filename, no path, no extension'
        
    def run(self):
        p = self.inputs['file_path']
        dir_path = os.path.split(p)[0]
        file_nopath = os.path.split(p)[1]
        file_noext = os.path.splitext(file_nopath)[0]
        self.outputs['dir_path'] = dir_path 
        self.outputs['filename'] = file_noext 
        self.message_callback('reading {}'.format(p))
        self.outputs['FabioImage'] = fabio.open(p)
        self.outputs['image_data'] = self.outputs['FabioImage'].data
        return self.outputs




import os
from collections import OrderedDict

from ..Workflow import Workflow 
from ...operations.IO.IMAGE.FabIOOpen import FabIOOpen
from ...operations.PROCESSING.INTEGRATION.Integrate1d import Integrate1d
from ...operations.PROCESSING.ZINGERS.EasyZingers1d import EasyZingers1d
from ...operations.IO.NumpySave import NumpySave

inputs = OrderedDict(
    image_file = None,
    image_data = None,
    calib_file = None,
    integrator = None,
    q_min = 0.,
    q_max = float('inf'),
    n_points = 1000.,
    polz_factor = 0.,
    output_dir = None,
    output_filename = None
    )

outputs = OrderedDict(
    q_I = None,
    q_I_dz = None,
    q_I_file = None,
    q_I_dz_file = None
    )

class Integrate(Workflow):

    def __init__(self):
        super(Integrate,self).__init__(inputs,outputs)
        self.add_operations(
            read_image=FabIOOpen(),
            integrate=Integrate1d(),
            dezinger=EasyZingers1d(),
            write_q_I=NumpySave(),
            write_q_I_dz=NumpySave()
            )

    def run(self):
        img_data = self.inputs['image_data'] 
        fn = self.inputs['output_filename'] 
        if self.inputs['image_file']:
            self.operations['read_image'].run_with(file_path=self.inputs['image_file'])
            img_data = self.operations['read_image'].outputs['image_data']
            if not fn: 
                fn = os.path.splitext(os.path.split(self.inputs['image_file'])[1])[0]
        intgtr = self.inputs['integrator']
        if self.inputs['calib_file'] is not None:
            intgtr.set_calib_file(self.inputs['calib_file'])
        if not intgtr.running:
            intgtr.start()
        self.operations['integrate'].run_with(
            integrator = intgtr,
            image_data = img_data,
            npt = self.inputs['n_points'],
            polz_factor = self.inputs['polz_factor']
            )
        all_q_I = self.operations['integrate'].outputs['q_I']
        q_min = self.inputs['q_min']
        q_max = self.inputs['q_max']
        idx_keep = ((all_q_I[:,0] >= q_min) & (all_q_I[:,0] <= q_max))
        q_I = all_q_I[idx_keep,:] 
        self.outputs['q_I'] = q_I 
        q_I_file = os.path.join(self.inputs['output_dir'],fn+'.dat')
        self.outputs['q_I_file'] = q_I_file
        if self.ops_enabled['write_q_I']:
            self.operations['write_q_I'].run_with(
                data = q_I,
                file_path = q_I_file,
                header = 'q (1/Angstrom), Intensity (arb)'
                )

        if self.ops_enabled['dezinger']:
            self.operations['dezinger'].run_with(q_I = q_I)
            q_I_dz = self.operations['dezinger'].outputs['q_I_dz']
            self.outputs['q_I_dz'] = q_I_dz
            q_I_dz_file = os.path.join(self.inputs['output_dir'],fn+'_dz.dat')
            self.outputs['q_I_dz_file'] = q_I_dz_file
            if self.ops_enabled['write_q_I_dz']:
                self.operations['write_q_I_dz'].run_with(
                    data = q_I_dz,
                    file_path = q_I_dz_file,
                    header = 'q (1/Angstrom), Intensity (arb)'
                    )

        return self.outputs




import os

def stack_contains(itm,stk):
    for lst in stk:
        if itm in lst:
            return True
    return False

def stack_size(stk):
    sz = 0
    for lst in stk:
        sz += len(lst)
    return sz

def print_stack(stk):
    stktxt = ''
    opt_newline = os.linesep
    n_layers = len(stk)
    for i,lst in enumerate(stk):
        if i == n_layers-1:
            opt_newline = ''
        stktxt += ('{}'+opt_newline).format(lst)
    return stktxt






from __future__ import print_function
from collections import OrderedDict
import socket 
from socket import AF_INET, SOCK_STREAM
from threading import Condition

from .PawsPlugin import PawsPlugin

content = OrderedDict(
    host=None,
    port=None,
    timer=None,
    mar_root_dir='/',
    verbose=False)

class MarCCDClient(PawsPlugin):

    STATE_MASK_BUSY = int(0x8)
    STATE_MASK_ACQUIRING = int(0x30)
    STATE_MASK_READING = int(0x300)
    STATE_MASK_CORRECTING = int(0x3000)
    STATE_MASK_WRITING = int(0x30000)
    STATE_MASK_DEZINGERING = int(0x300000)
    STATE_MASK_SAVING = int(0x33300)
    STATE_MASK_ERROR = int(0x44444)
    #TIMEOUT = 60
    #URGENT_TIMEOUT = 0.5
    #BASE_IMAGE_SIZE = 4096

    def __init__(self):
        super(MarCCDClient,self).__init__(content)
        self.content_doc['host'] = 'string representing host name or IP address'
        self.content_doc['port'] = 'integer port number where Mar server listens' 
        self.content_doc['timer'] = 'Timer plugin for triggering client activities' 
        self.content_doc['verbose'] = 'If True, plugin uses its message_callback' 
        # this lock allows Operations to wait for exposure loops to finish
        self.exposure_lock = Condition()
        self.sock = None
        self.thread_blocking = True
        self.state_lock = Condition()
        self.state = OrderedDict(
            state_code = None
            )

    def description(self):
        return 'MarCCDClient Plugin: '\
            'TCP/IP Client used to communicate with MarCCD server'

    def start(self):
        super(MarCCDClient,self).start() 

    def run(self):
        hst = self.content['host'] 
        prt = self.content['port'] 
        self.sock = socket.create_connection((hst,prt)) 
        self.vb = self.content['verbose']
        tmr = self.content['timer'] 
        self.mar_root = self.content['mar_root_dir']

        self.set_darkfield()       

        self.update_status()
        keep_going = True
        while keep_going: 
            with tmr.dt_lock:
                tmr.dt_lock.wait()
            while self.commands.qsize() > 0:
                with self.command_lock:
                    cmd = self.commands.get()
                    self.commands.task_done()
                cmd_parts = cmd.split(',')
                if cmd_parts[0] == 'expose':
                    self.expose(float(cmd_parts[1]),cmd_parts[2])
                #self.run_cmd(cmd)
            self.update_status()
            with tmr.running_lock:
                if not tmr.running:
                    with self.proxy.running_lock:
                        self.proxy.stop()
            with self.proxy.running_lock:
                keep_going = bool(self.proxy.running)
        with self.proxy.history_lock:
            self.proxy.add_to_history(t_now,cmd,resp.decode())
            self.proxy.dump_history()
        self.sock.close()

    def expose(self,integration_time,filename):
        self.wait_for_state([self.STATE_MASK_ACQUIRING,self.STATE_MASK_BUSY,self.STATE_MASK_READING],False)
        if self.vb: self.message_callback('starting {}-second exposure'.format(integration_time))
        self.run_cmd('start') 
        self.wait_for_state(self.STATE_MASK_ACQUIRING)
        self.send_cmd('shutter,0')
        time.sleep(integration_time)
        self.send_cmd('shutter,1')
        if self.vb: self.message_callback('saving exposure to register 0')
        self.run_cmd('readout,0')
        # NOTE: have to expose twice
        # into two different registers (2 and 0)
        # for dezinger filter to work
        #self.run_cmd('dezinger,0')
        if self.vb: self.message_callback('correcting exposure')
        self.run_cmd('correct')
        if self.vb: self.message_callback('writing exposure to {}'.format(self.mar_root+filename))
        self.run_cmd('writefile,{},1'.format(self.mar_root+filename))

    def set_darkfield(self,integration_time=0.):
        if self.vb: self.message_callback('starting darkfield scan to register 2')
        self.run_cmd('start')
        self.wait_for_state(self.STATE_MASK_ACQUIRING)
        time.sleep(integration_time)
        self.run_cmd('readout,2')
        self.wait_for_idle()
        if self.vb: self.message_callback('starting darkfield scan to register 1')
        self.run_cmd('start')
        self.wait_for_state(self.STATE_MASK_ACQUIRING)
        time.sleep(integration_time)
        self.run_cmd('readout,1')
        self.wait_for_state([self.STATE_MASK_ACQUIRING,self.STATE_MASK_BUSY,self.STATE_MASK_READING],False)
        if self.vb: self.message_callback('starting two-image dezinger')
        self.run_cmd('dezinger,1')
        self.wait_for_state([self.STATE_MASK_BUSY,self.STATE_MASK_DEZINGERING],False)

    def wait_for_idle(self,additional_delay=1.3):
        if self.content['verbose']: self.message_callback('waiting for Mar server to be idle')
        while self.state['state_code'] in [\
            self.STATE_MASK_ACQUIRING, self.STATE_MASK_READING,\
            self.STATE_MASK_CORRECTING, self.STATE_MASK_WRITING,\
            self.STATE_MASK_DEZINGERING, self.STATE_MASK_BUSY]:
            self.update_status()
        # an additional delay is implemented to allow the server to truly become idle.
        # in py4syn, this delay was empirically tuned to 1.3 seconds 
        # for reliable two-image acquisition.
        time.sleep(additional_delay) 

    def wait_for_state(self,state,value=True):
        if not isinstance(state,list): state = [state]
        if value:
            while not self.state['state_code'] in state:
                self.update_status()
        else:
            while self.state['state_code'] in state:
                self.update_status()

    def update_status(self):
        with self.state_lock:
            self.read_status()
        with self.proxy.state_lock:
            self.proxy.state = copy.deepcopy(self.state)
        if vb: self.message_callback(self.print_status())

    def read_status(self):
        resp = self.run_cmd('get_state')
        self.state['state_code'] = int(resp)

    def run_cmd(self,cmd):
        with self.content['timer'].dt_lock:
            t_now = float(self.content['timer'].dt_utc())
        self.send_line(cmd)
        resp = self.receive_line()
        with self.proxy.history_lock:
            self.proxy.add_to_history(t_now,cmd+' '+resp)
        return resp

    def receive_line(self):
        bfr = bytearray(b' ' * 1024) 
        ln = self.sock.recv_into(bfr) 
        return bfr.strip().strip(b'\x00\n').decode()
    
    def send_line(self, line):
        self.sock.sendall(bytearray(line.encode('utf-8')))

    def request_exposure(self,integration_time,filename):
        with self.thread_clone.command_lock:
            self.thread_clone.commands.put('expose,{},{}'.format(integration_time,filename))





import os
import copy
from collections import OrderedDict

import numpy as np

from ....Workflow import Workflow 
from ...BL15.LEGACY.ReadTimeSeries import ReadTimeSeries 
from .....operations.PACKAGING.PIF.PackXRSDPIF import PackXRSDPIF
from .....operations.IO.CITRINATION.UploadPIF import UploadPIF

inputs = OrderedDict(
    citrination_client = None,
    source_wavelength = None,
    header_dir = None,
    header_suffix = None,
    header_regex = None,
    q_I_dir = None,
    q_I_suffix = '',
    q_I_ext = '.dat',
    system_dir = None, 
    system_suffix = '',
    dataset_id = None,
    upload_data = False,
    upload_pif = False,
    keep_pif_json = True
    )

outputs = OrderedDict()

class BatchXRSDPIF(Workflow):
    
    def __init__(self):
        super(BatchXRSDPIF,self).__init__(inputs,outputs)
        self.add_operations(
            read = ReadTimeSeries(),
            make_pif = PackXRSDPIF(),
            upload_pif = UploadPIF()
            )

    def run(self):
        src_wl = self.inputs['source_wavelength']
        # READ THE TIMESERIES 
        self.operations['read'].operations['read_batch'].operations['read'].disable_ops('read_image')
        ts_data = self.operations['read'].run_with(
            header_dir = self.inputs['header_dir'],
            header_suffix = self.inputs['header_suffix'],
            header_regex = self.inputs['header_regex'],
            q_I_dir = self.inputs['q_I_dir'], 
            q_I_suffix = self.inputs['q_I_suffix'],
            q_I_ext = self.inputs['q_I_ext'],
            system_dir = self.inputs['system_dir'], 
            system_suffix = self.inputs['system_suffix'] 
            )
        # CREATE TIMESERIES OF PIFS
        all_pifs = []
        ts = ts_data['time']
        hdrs = ts_data['header_data']
        syss = ts_data['system']
        q_Is = ts_data['q_I']
        q_I_files = ts_data['q_I_files']
        npifs = len(ts)
        for ipif, t_utc, hdr_data, sys, q_I, q_I_file in zip(range(len(ts)),ts,hdrs,syss,q_Is,q_I_files): 
            self.message_callback('making PIF {} / {} ({})'.format(ipif,npifs,sys.sample_metadata['sample_id']))
            pif_outputs = self.operations['make_pif'].run_with(
                system = sys,
                q_I = None, 
                q_I_file = os.path.split(q_I_file)[1], 
                t_utc = t_utc,
                temperature = hdr_data['temperature'],
                source_wavelength = src_wl
                )
            all_pifs.append(pif_outputs['pif'])
        # UPLOAD THE q_I DATA FILES
        if self.inputs['upload_data']:
            for ifp, fp in enumerate(q_I_files):
                fn = os.path.split(fp)[1]
                self.message_callback('Uploading data file {} / {} ({})'.format(ifp,npifs,fn))
                self.inputs['citrination_client'].client.upload(self.inputs['dataset_id'],fp,fn)
        # SAVE AND UPLOAD THE PIFS
        pif_json_path = os.path.join(self.inputs['system_dir'],'pif.json') 
        self.operations['upload_pif'].run_with(
            pif = all_pifs,
            citrination_client = self.inputs['citrination_client'],
            dsid = self.inputs['dataset_id'],
            json_path = pif_json_path,
            keep_json = self.inputs['keep_pif_json'],
            upload_flag = self.inputs['upload_pif']
            )
        return self.outputs




from collections import OrderedDict
import time
import os

from ...Operation import Operation

inputs=OrderedDict(
    spec_infoclient=None,
    T_set=None,
    T_ramp=None,
    delay_time=0)
outputs=OrderedDict(
    report=None)
        
class SetCryoCon(Operation):
    """Set the temperature of the CryoCon"""

    def __init__(self):
        super(SetCryoCon,self).__init__(inputs,outputs)
        self.input_doc['spec_infoclient'] = 'A SpecInfoClient'
        self.input_doc['T_set'] = 'Desired setpoint in degrees C' 
        self.input_doc['T_ramp'] = 'Desired temperature ramp in degrees C per minute' 
        self.input_doc['delay_time'] = 'Seconds to sleep after setting the cryocon' 
        self.output_doc['report'] = 'dict reporting details of final state' 

    def run(self):
        cl = self.inputs['spec_infoclient'] 
        T_set = self.inputs['T_set'] 
        T_ramp = self.inputs['T_ramp'] 
        delay = self.inputs['delay_time']
        self.message_callback('enabling cryocon')
        cl.enable_cryocon()
        self.message_callback('setting cryocon to {}, at {} degC/minute'.format(T_set,T_ramp))
        cl.set_cryocon(T_set,T_ramp)
        if delay:
            self.message_callback('delaying {} seconds...'.format(delay))
            time.sleep(delay)
            # TODO: implement a control to ensure the setpoint
        val = cl.read_cryocon()
        rpt = dict(targets=[T_set],final_values=[val])
        self.outputs['report'] = rpt
        #self.outputs['flag'] = True
        #else:
        #    self.outputs['report'] = {'STATUS':bool(stat)}
        #    self.outputs['flag'] = False 

